`timescale 1ns / 1ps

module ML505Top(
    GPIO_SW_C,
    USER_CLK,

    FPGA_SERIAL_RX0,
    FPGA_SERIAL_TX0,

    FPGA_SERIAL_RX1,
    FPGA_SERIAL_TX1
);

input                           GPIO_SW_C;
input                           USER_CLK;

input                           FPGA_SERIAL_RX0;
output                          FPGA_SERIAL_TX0;

input                           FPGA_SERIAL_RX1;
output                          FPGA_SERIAL_TX1;

// CPU
wire        [31:0]              fetchaddress_from_cpu;
wire                            fetchenable_from_cpu;
wire        [31:0]              fetchout_from_imem;

wire        [31:0]              dataaddress_from_cpu;
wire        [31:0]              cpudatain_from_mmap;
wire                            cpudatainvalid_from_mmap;
wire                            datainenable_from_cpu;
wire        [31:0]              dataout_from_cpu;
wire                            dataoutenable_from_cpu;
wire                            cpudataoutready_from_mmap;

reg         [31:0]              prevdataout_from_cpu;

// COP0
wire        [4:0]               cop0address_from_cpu;
wire        [31:0]              dataout_from_cop0;
wire                            cop0datainenable_from_cpu;
wire        [31:0]              cop0datain_from_cpu;

wire        [31:0]              cop0interruptedpc_from_cpu;
wire                            cop0interrupthandled_from_cpu;
wire                            interruptrequest_from_cop0;

// Instruction Memory
wire        [15:0]              imemfetchaddress_from_mmap;
wire                            imemfetchenable_from_mmap;

wire        [15:0]              imemdataaddress_from_mmap;
wire                            imemdatainenable_from_mmap;
wire                            datainready_from_imem;
wire        [31:0]              dataout_from_imem;
wire                            dataoutvalid_from_imem;
wire                            imemdataoutenable_from_mmap;

// Data Memory
wire        [15:0]              dmemdataaddress_from_mmap;
wire                            dmemdatainenable_from_mmap;
wire                            datainready_from_dmem;
wire        [31:0]              dataout_from_dmem;
wire                            dataoutvalid_from_dmem;
wire                            dmemdataoutenable_from_mmap;

// UART0
wire                            uart0datainvalid_from_mmap;
wire                            datainready_from_uart0;
wire        [7:0]               dataout_from_uart0;
wire                            dataoutvalid_from_uart0;
wire                            uart0dataoutready_from_mmap;
wire                            uart0irq;

reg                             prevdatainready_from_uart0;
reg                             prevdataoutvalid_from_uart0;

wire                            uart1datainvalid_from_mmap;
wire                            datainready_from_uart1;
wire        [7:0]               dataout_from_uart1;
wire                            dataoutvalid_from_uart1;
wire                            uart1dataoutready_from_mmap;
wire                            uart1irq;

reg                             prevdatainready_from_uart1;
reg                             prevdataoutvalid_from_uart1;

wire                            cpureset;

wire                            clock125mhz;


always @(posedge clock125mhz)
begin
    prevdataout_from_cpu <= dataout_from_cpu;

    prevdatainready_from_uart0 <= datainready_from_uart0;
    prevdataoutvalid_from_uart0 <= dataoutvalid_from_uart0;

    prevdatainready_from_uart1 <= datainready_from_uart1;
    prevdataoutvalid_from_uart1 <= dataoutvalid_from_uart1;
end

SourceOfClocks clocksource(
    .Clock100MHz(               USER_CLK),
    .Clock125MHz(               clock125mhz)
);

ResetControl resetcontrol(
    .Clock(                     clock125mhz),
    .CPUTrigger(                GPIO_SW_C),
    .CPUReset(                  cpureset)
);

APMIPS smips150(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),
    .Reset(                     cpureset),

    .FetchAddress(              fetchaddress_from_cpu),
    .FetchEnable(               fetchenable_from_cpu),
    .FetchIn(                   fetchout_from_imem),

    .DataAddress(               dataaddress_from_cpu),
    .DataIn(                    cpudatain_from_mmap),
    .DataInEnable(              datainenable_from_cpu),
    .DataOut(                   dataout_from_cpu),
    .DataOutEnable(             dataoutenable_from_cpu),

    .COP0DataAddress(           cop0address_from_cpu),
    .COP0DataIn(                dataout_from_cop0),
    .COP0DataOut(               cop0datain_from_cpu),
    .COP0DataOutEnable(         cop0datainenable_from_cpu),

    .InterruptedPC(             cop0interruptedpc_from_cpu),
    .InterruptHandled(          cop0interrupthandled_from_cpu),
    .InterruptRequest(          interruptrequest_from_cop0)
);

assign uart0irq = (datainready_from_uart0 & ~prevdatainready_from_uart0)
                | (dataoutvalid_from_uart0 & ~prevdataoutvalid_from_uart0);

assign uart1irq = (datainready_from_uart1 & ~prevdatainready_from_uart1)
                | (dataoutvalid_from_uart1 & ~prevdataoutvalid_from_uart1);

COP0150 cop0150(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),
    .Reset(                     cpureset),

    .DataAddress(               cop0address_from_cpu),
    .DataOut(                   dataout_from_cop0),
    .DataInEnable(              cop0datainenable_from_cpu),
    .DataIn(                    cop0datain_from_cpu),

    .InterruptedPC(             cop0interruptedpc_from_cpu),
    .InterruptHandled(          cop0interrupthandled_from_cpu),
    .InterruptRequest(          interruptrequest_from_cop0),

    .UART0Request(              uart0irq),
    .UART1Request(              uart1irq)
);

MemoryMap #(
    .InstructionStartAddress(   32'h0040_0000),
    .InstructionAddressWidth(   16),

    .DataStartAddress(          32'h1001_0000),
    .DataAddressWidth(          16),

    .UART0ControlInAddress(     32'hFFFF_0000),
    .UART0DataInAddress(        32'hFFFF_0004),
    .UART0ControlOutAddress(    32'hFFFF_0008),
    .UART0DataOutAddress(       32'hFFFF_000C),

    .UART1ControlInAddress(     32'hFFFF_0010),
    .UART1DataInAddress(        32'hFFFF_0014),
    .UART1ControlOutAddress(    32'hFFFF_0018),
    .UART1DataOutAddress(       32'hFFFF_001C)
) mmap(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),

    // CPU
    .FetchAddressFromCPU(       fetchaddress_from_cpu),
    .FetchEnableFromCPU(        fetchenable_from_cpu),

    .DataAddressFromCPU(        dataaddress_from_cpu),
    .CPUDataIn(                 cpudatain_from_mmap),
    .DataInEnableFromCPU(       datainenable_from_cpu),
    .DataOutEnableFromCPU(      dataoutenable_from_cpu),

    // Instruction Memory
    .IMemFetchAddress(          imemfetchaddress_from_mmap),
    .IMemFetchEnable(           imemfetchenable_from_mmap),

    .IMemDataAddress(           imemdataaddress_from_mmap),
    .IMemDataInEnable(          imemdatainenable_from_mmap),
    .DataOutFromIMem(           dataout_from_imem),
    .IMemDataOutEnable(         imemdataoutenable_from_mmap),

    // Data Memory
    .DMemDataAddress(           dmemdataaddress_from_mmap),
    .DMemDataInEnable(          dmemdatainenable_from_mmap),
    .DataOutFromDMem(           dataout_from_dmem),
    .DMemDataOutEnable(         dmemdataoutenable_from_mmap),

    // UART0
    .UART0DataInValid(           uart0datainvalid_from_mmap),
    .DataInReadyFromUART0(       datainready_from_uart0),
    .DataOutFromUART0(           dataout_from_uart0),
    .DataOutValidFromUART0(      dataoutvalid_from_uart0),
    .UART0DataOutReady(          uart0dataoutready_from_mmap),

    // UART1
    .UART1DataInValid(           uart1datainvalid_from_mmap),
    .DataInReadyFromUART1(       datainready_from_uart1),
    .DataOutFromUART1(           dataout_from_uart1),
    .DataOutValidFromUART1(      dataoutvalid_from_uart1),
    .UART1DataOutReady(          uart1dataoutready_from_mmap)
);


// Instruction Memory
InstructionMemory #(
    .DataWidth(                 32),
    .AddressWidth(              16),
    .InitialHexFile(            "")
) imem(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),

    .FetchAddress(              imemfetchaddress_from_mmap),
    .FetchOut(                  fetchout_from_imem),
    .FetchEnable(               imemfetchenable_from_mmap),

    .DataAddress(               imemdataaddress_from_mmap),
    .DataIn(                    dataout_from_cpu),
    .DataInEnable(              imemdatainenable_from_mmap),
    .DataOut(                   dataout_from_imem),
    .DataOutEnable(             imemdataoutenable_from_mmap)
);

// Data Memory
DataMemory #(
    .DataWidth(                 32),
    .AddressWidth(              16),
    .InitialHexFile(            "")
) dmem(
    .Clock(                     clock125mhz),
    .Enable(                    1'b1),

    .DataAddress(               dmemdataaddress_from_mmap),
    .DataIn(                    dataout_from_cpu),
    .DataInEnable(              dmemdatainenable_from_mmap),
    .DataOut(                   dataout_from_dmem),
    .DataOutEnable(             dmemdataoutenable_from_mmap)
);

// UART0
UART #(
    .ClockFreq(                 100_000_000),
    .Baud(                      115_200),
    .Width(                     8),
    .Parity(                    0),
    .StopBits(                  1)
) uart0(
    .Clock(                     clock125mhz),
    .Reset(                     cpureset),

    .DataIn(                    prevdataout_from_cpu[7:0]),
    .DataInValid(               uart0datainvalid_from_mmap),
    .DataInReady(               datainready_from_uart0),

    .DataOut(                   dataout_from_uart0),
    .DataOutValid(              dataoutvalid_from_uart0),
    .DataOutReady(              uart0dataoutready_from_mmap),

    .SIn(                       FPGA_SERIAL_RX0),
    .SOut(                      FPGA_SERIAL_TX0)
);

// UART1
UART #(
    .ClockFreq(                 100_000_000),
    .Baud(                      115_200),
    .Width(                     8),
    .Parity(                    0),
    .StopBits(                  1)
) uart1(
    .Clock(                     clock125mhz),
    .Reset(                     cpureset),

    .DataIn(                    prevdataout_from_cpu[7:0]),
    .DataInValid(               uart1datainvalid_from_mmap),
    .DataInReady(               datainready_from_uart1),

    .DataOut(                   dataout_from_uart1),
    .DataOutValid(              dataoutvalid_from_uart1),
    .DataOutReady(              uart1dataoutready_from_mmap),

    .SIn(                       FPGA_SERIAL_RX1),
    .SOut(                      FPGA_SERIAL_TX1)
);

endmodule
