module MemoryMap(
    Clock,
    Enable,

    // CPU
    FetchAddressFromCPU,
    FetchEnableFromCPU,

    DataAddressFromCPU,
    CPUDataIn,
    DataInEnableFromCPU,
    DataOutEnableFromCPU,

    // Instruction Memory
    IMemFetchAddress,
    IMemFetchEnable,

    IMemDataAddress,
    IMemDataInEnable,
    DataOutFromIMem,
    IMemDataOutEnable,

    // Data Memory
    DMemDataAddress,
    DMemDataInEnable,
    DataOutFromDMem,
    DMemDataOutEnable,

    // UART0
    UART0DataInValid,
    DataInReadyFromUART0,
    DataOutFromUART0,
    DataOutValidFromUART0,
    UART0DataOutReady,

    // UART1
    UART1DataInValid,
    DataInReadyFromUART1,
    DataOutFromUART1,
    DataOutValidFromUART1,
    UART1DataOutReady
);

parameter InstructionStartAddress = 32'h0040_0000,
          InstructionAddressWidth = 16,
          DataStartAddress = 32'h1001_0000,
          DataAddressWidth = 16,
          UART0ControlInAddress = 32'hFFFF_0000,
          UART0DataInAddress = 32'hFFFF_0004,
          UART0ControlOutAddress = 32'hFFFF_0008,
          UART0DataOutAddress = 32'hFFFF_000C,
          UART1ControlInAddress = 32'hFFFF_0010,
          UART1DataInAddress = 32'hFFFF_0014,
          UART1ControlOutAddress = 32'hFFFF_0018,
          UART1DataOutAddress = 32'hFFFF_001C;

localparam IAddrWidthM1 = InstructionAddressWidth-1,
           DAddrWidthM1 = DataAddressWidth-1;

input                           Clock;
input                           Enable;

input       [31:0]              FetchAddressFromCPU;
input                           FetchEnableFromCPU;

input       [31:0]              DataAddressFromCPU;
output      [31:0]              CPUDataIn;
input                           DataInEnableFromCPU;
input                           DataOutEnableFromCPU;

output      [IAddrWidthM1:0]    IMemFetchAddress;
output                          IMemFetchEnable;

output      [IAddrWidthM1:0]    IMemDataAddress;
output                          IMemDataInEnable;
input       [31:0]              DataOutFromIMem;
output                          IMemDataOutEnable;

output      [DAddrWidthM1:0]    DMemDataAddress;
output                          DMemDataInEnable;
input       [31:0]              DataOutFromDMem;
output                          DMemDataOutEnable;

output                          UART0DataInValid;
input                           DataInReadyFromUART0;
input       [7:0]               DataOutFromUART0;
input                           DataOutValidFromUART0;
output                          UART0DataOutReady;

output                          UART1DataInValid;
input                           DataInReadyFromUART1;
input       [7:0]               DataOutFromUART1;
input                           DataOutValidFromUART1;
output                          UART1DataOutReady;

reg         [31:0]              dataaddress_from_cpu;    
reg                             datainready_from_cpu;
reg                             dataoutvalid_from_cpu;

reg         [31:0]              cpudatain;

// CPU
assign CPUDataIn = cpudatain;

// Instruction Memory
assign IMemFetchAddress = FetchAddressFromCPU - InstructionStartAddress;
assign IMemFetchEnable  = FetchEnableFromCPU && FetchAddressFromCPU >= InstructionStartAddress &&
                                                FetchAddressFromCPU < (InstructionStartAddress + 2 ** InstructionAddressWidth);

assign IMemDataAddress   = DataAddressFromCPU - InstructionStartAddress;
assign IMemDataOutEnable = DataInEnableFromCPU && DataAddressFromCPU >= InstructionStartAddress &&
                                                  DataAddressFromCPU < (InstructionStartAddress + 2 ** InstructionAddressWidth);
assign IMemDataInEnable  = DataOutEnableFromCPU && DataAddressFromCPU >= InstructionStartAddress &&
                                                   DataAddressFromCPU < (InstructionStartAddress + 2 ** InstructionAddressWidth);

// Data Memory
assign DMemDataAddress   = DataAddressFromCPU - DataStartAddress;
assign DMemDataOutEnable = DataInEnableFromCPU && DataAddressFromCPU >= DataStartAddress &&
                                                  DataAddressFromCPU < (DataStartAddress + 2 ** DataAddressWidth);
assign DMemDataInEnable  = DataOutEnableFromCPU && DataAddressFromCPU >= DataStartAddress &&
                                                   DataAddressFromCPU < (DataStartAddress + 2 ** DataAddressWidth);

// UART0
assign UART0DataOutReady = datainready_from_cpu && dataaddress_from_cpu == UART0DataInAddress;
assign UART0DataInValid  = dataoutvalid_from_cpu && dataaddress_from_cpu == UART0DataOutAddress;

// UART1
assign UART1DataOutReady = datainready_from_cpu && dataaddress_from_cpu == UART1DataInAddress;
assign UART1DataInValid  = dataoutvalid_from_cpu && dataaddress_from_cpu == UART1DataOutAddress;

always@(posedge Clock) begin
    dataaddress_from_cpu  <= DataAddressFromCPU;
    datainready_from_cpu  <= DataInEnableFromCPU;
    dataoutvalid_from_cpu <= DataOutEnableFromCPU;
end

always@(*) begin
    if(datainready_from_cpu) begin
        if(dataaddress_from_cpu >= InstructionStartAddress &&
           dataaddress_from_cpu < (InstructionStartAddress + 2 ** InstructionAddressWidth)) begin
            cpudatain       = DataOutFromIMem;
        end else if(dataaddress_from_cpu >= DataStartAddress &&
                    dataaddress_from_cpu < (DataStartAddress + 2 ** DataAddressWidth)) begin
            cpudatain       = DataOutFromDMem;
        end else begin
            case(dataaddress_from_cpu)
                UART0ControlInAddress: begin
                    cpudatain       = {31'b0, DataOutValidFromUART0};
                end
                UART0DataInAddress: begin
                    cpudatain       = {24'b0, DataOutFromUART0};
                end
                UART0ControlOutAddress: begin
                    cpudatain       = {31'b0, DataInReadyFromUART0};
                end
                UART1ControlInAddress: begin
                    cpudatain       = {31'b0, DataOutValidFromUART1};
                end
                UART1DataInAddress: begin
                    cpudatain       = {24'b0, DataOutFromUART1};
                end
                UART1ControlOutAddress: begin
                    cpudatain       = {31'b0, DataInReadyFromUART1};
                end
                default: begin
                    cpudatain       = 32'bx;
                end
            endcase
        end
    end else begin
        cpudatain       = 32'bx;
    end
end

endmodule
